@@ -85,8 +85,12 @@ void s_poll_add(s_poll_set *fds, int fd, int rd, int wr) {
         fds->ufds[i].events=0;
         fds->nfds++;
     }
-    if(rd)
+    if(rd) {
         fds->ufds[i].events|=POLLIN;
+#ifdef POLLRDHUP
+        fds->ufds[i].events|=POLLRDHUP;
+#endif
+    }
     if(wr)
         fds->ufds[i].events|=POLLOUT;
 }
@@ -109,12 +113,27 @@ int s_poll_canwrite(s_poll_set *fds, int fd) {
     return 0; /* not listed in fds */
 }
 
+/* best doc: http://lxr.free-electrons.com/source/net/ipv4/tcp.c#L456 */
+
 int s_poll_hup(s_poll_set *fds, int fd) {
     unsigned int i;
 
     for(i=0; i<fds->nfds; i++)
         if(fds->ufds[i].fd==fd)
-            return fds->ufds[i].revents&POLLHUP;
+            return fds->ufds[i].revents&POLLHUP; /* read and write closed */
+    return 0; /* not listed in fds */
+}
+
+int s_poll_rdhup(s_poll_set *fds, int fd) {
+    unsigned int i;
+
+    for(i=0; i<fds->nfds; i++)
+        if(fds->ufds[i].fd==fd)
+#ifdef POLLRDHUP
+            return fds->ufds[i].revents&POLLRDHUP; /* read closed */
+#else
+            return fds->ufds[i].revents&POLLHUP; /* read and write closed */
+#endif
     return 0; /* not listed in fds */
 }
 
@@ -370,6 +389,12 @@ int s_poll_hup(s_poll_set *fds, int fd) {
     return 0; /* FIXME: how to detect HUP condition with select()? */
 }
 
+int s_poll_rdhup(s_poll_set *fds, int fd) {
+    (void)fds; /* skip warning about unused parameter */
+    (void)fd; /* skip warning about unused parameter */
+    return 0; /* FIXME: how to detect RDHUP condition with select()? */
+}
+
 int s_poll_error(s_poll_set *fds, int fd) {
     /* error conditions are signaled as read, but apparently *not* in Winsock:
      * http://msdn.microsoft.com/en-us/library/windows/desktop/ms737625%28v=vs.85%29.aspx */
@@ -553,8 +578,7 @@ void s_write(CLI *c, int fd, const void *buf, int len) {
             longjmp(c->err, 1); /* error */
         }
         num=writesocket(fd, (void *)ptr, len);
-        switch(num) {
-        case -1: /* error */
+        if(num==-1) { /* error */
             sockerror("writesocket (s_write)");
             longjmp(c->err, 1);
         }
@@ -618,50 +642,23 @@ char *fd_getline(CLI *c, int fd) {
 
     line=str_alloc(allocated);
     for(;;) {
-        s_poll_init(c->fds);
-        s_poll_add(c->fds, fd, 1, 0); /* read */
-        switch(s_poll_wait(c->fds, c->opt->timeout_busy, 0)) {
-        case -1:
-            sockerror("fd_getline: s_poll_wait");
+        if(ptr>65536) { /* >64KB --> DoS protection */
+            s_log(LOG_ERR, "fd_getline: Line too long");
             str_free(line);
-            longjmp(c->err, 1); /* error */
-        case 0:
-            s_log(LOG_INFO, "fd_getline: s_poll_wait:"
-                " TIMEOUTbusy exceeded: sending reset");
-            str_free(line);
-            longjmp(c->err, 1); /* timeout */
-        case 1:
-            break; /* OK */
-        default:
-            s_log(LOG_ERR, "fd_getline: s_poll_wait: Unknown result");
-            str_free(line);
-            longjmp(c->err, 1); /* error */
+            longjmp(c->err, 1);
         }
         if(allocated<ptr+1) {
             allocated*=2;
             line=str_realloc(line, allocated);
         }
-        switch(readsocket(fd, line+ptr, 1)) {
-        case -1: /* error */
-            sockerror("fd_getline: readsocket");
-            str_free(line);
-            longjmp(c->err, 1);
-        case 0: /* EOF */
-            s_log(LOG_ERR, "fd_getline: Unexpected socket close");
-            str_free(line);
-            longjmp(c->err, 1);
-        }
+        s_read(c, fd, line+ptr, 1);
         if(line[ptr]=='\r')
             continue;
         if(line[ptr]=='\n')
             break;
         if(line[ptr]=='\0')
             break;
-        if(++ptr>65536) { /* >64KB --> DoS protection */
-            s_log(LOG_ERR, "fd_getline: Line too long");
-            str_free(line);
-            longjmp(c->err, 1);
-        }
+        ++ptr;
     }
     line[ptr]='\0';
     tmpline=str_dup(line);
@@ -686,6 +683,98 @@ void fd_printf(CLI *c, int fd, const char *format, ...) {
     str_free(line);
 }
 
+void s_ssl_write(CLI *c, const void *buf, int len) {
+        /* simulate a blocking SSL_write */
+    u8 *ptr=(u8 *)buf;
+    int num;
+
+    while(len>0) {
+        s_poll_init(c->fds);
+        s_poll_add(c->fds, c->ssl_wfd->fd, 0, 1); /* write */
+        switch(s_poll_wait(c->fds, c->opt->timeout_busy, 0)) {
+        case -1:
+            sockerror("s_write: s_poll_wait");
+            longjmp(c->err, 1); /* error */
+        case 0:
+            s_log(LOG_INFO, "s_write: s_poll_wait:"
+                " TIMEOUTbusy exceeded: sending reset");
+            longjmp(c->err, 1); /* timeout */
+        case 1:
+            break; /* OK */
+        default:
+            s_log(LOG_ERR, "s_write: s_poll_wait: unknown result");
+            longjmp(c->err, 1); /* error */
+        }
+        num=SSL_write(c->ssl, (void *)ptr, len);
+        if(num==-1) { /* error */
+            sockerror("SSL_write (s_ssl_write)");
+            longjmp(c->err, 1);
+        }
+        ptr+=num;
+        len-=num;
+    }
+}
+
+void s_ssl_read(CLI *c, void *ptr, int len) {
+        /* simulate a blocking SSL_read */
+    int num;
+
+    while(len>0) {
+        if(!SSL_pending(c->ssl)) {
+            s_poll_init(c->fds);
+            s_poll_add(c->fds, c->ssl_rfd->fd, 1, 0); /* read */
+            switch(s_poll_wait(c->fds, c->opt->timeout_busy, 0)) {
+            case -1:
+                sockerror("s_read: s_poll_wait");
+                longjmp(c->err, 1); /* error */
+            case 0:
+                s_log(LOG_INFO, "s_read: s_poll_wait:"
+                    " TIMEOUTbusy exceeded: sending reset");
+                longjmp(c->err, 1); /* timeout */
+            case 1:
+                break; /* OK */
+            default:
+                s_log(LOG_ERR, "s_read: s_poll_wait: unknown result");
+                longjmp(c->err, 1); /* error */
+            }
+        }
+        num=SSL_read(c->ssl, ptr, len);
+        switch(num) {
+        case -1: /* error */
+            sockerror("SSL_read (s_ssl_read)");
+            longjmp(c->err, 1);
+        case 0: /* EOF */
+            s_log(LOG_ERR, "Unexpected socket close (s_ssl_read)");
+            longjmp(c->err, 1);
+        }
+        ptr=(u8 *)ptr+num;
+        len-=num;
+    }
+}
+
+char *ssl_getstring(CLI *c) { /* get null-terminated string */
+    char *line;
+    int ptr=0, allocated=32;
+
+    line=str_alloc(allocated);
+    for(;;) {
+        if(ptr>65536) { /* >64KB --> DoS protection */
+            s_log(LOG_ERR, "fd_getline: Line too long");
+            str_free(line);
+            longjmp(c->err, 1);
+        }
+        if(allocated<ptr+1) {
+            allocated*=2;
+            line=str_realloc(line, allocated);
+        }
+        s_ssl_read(c, line+ptr, 1);
+        if(line[ptr]=='\0')
+            break;
+        ++ptr;
+    }
+    return line;
+}
+
     /* replace non-UTF-8 and non-printable control characters with '.' */
 void safestring(char *c) {
     for(; *c; ++c)
