@@ -177,7 +177,7 @@ typedef struct service_options_struct {
     char *cert;                                             /* cert filename */
     char *key;                               /* pem (priv key/cert) filename */
     long session_size, session_timeout;
-    long ssl_options;
+    long ssl_options_set, ssl_options_clear;
     SSL_METHOD *client_method, *server_method;
     SOCKADDR_UNION sessiond_addr;
 #ifndef OPENSSL_NO_TLSEXT
@@ -417,6 +417,7 @@ void s_poll_add(s_poll_set *, int, int, int);
 int s_poll_canread(s_poll_set *, int);
 int s_poll_canwrite(s_poll_set *, int);
 int s_poll_hup(s_poll_set *, int);
+int s_poll_rdhup(s_poll_set *, int);
 int s_poll_error(s_poll_set *, int);
 int s_poll_wait(s_poll_set *, int, int);
 
@@ -450,7 +451,7 @@ typedef struct {
     SOCKADDR_UNION peer_addr; /* peer address */
     socklen_t peer_addr_len;
     SOCKADDR_UNION *bind_addr; /* address to bind() the socket */
-    SOCKADDR_LIST connect_addr; /* for dynamically assigned addresses */
+    SOCKADDR_LIST connect_addr; /* either copied or resolved dynamically */
     FD local_rfd, local_wfd; /* read and write local descriptors */
     FD remote_fd; /* remote file descriptor */
         /* IP for explicit local bind or transparent proxy */
@@ -486,6 +487,9 @@ void fd_printf(CLI *, int, const char *, ...)
 #else
     ;
 #endif
+void s_ssl_write(CLI *, const void *, int);
+void s_ssl_read(CLI *, void *, int);
+char *ssl_getstring(CLI *c);
 
 /**************************************** prototype for protocol.c */
 
@@ -504,6 +508,8 @@ void resolver_init();
 int name2addr(SOCKADDR_UNION *, char *, char *);
 int hostport2addr(SOCKADDR_UNION *, char *, char *);
 int namelist2addrlist(SOCKADDR_LIST *, NAME_LIST *, char *);
+int hostport2addrlist(SOCKADDR_LIST *, char *, char *);
+void addrlist_init(SOCKADDR_LIST *);
 void addrlist_dup(SOCKADDR_LIST *, const SOCKADDR_LIST *);
 char *s_ntop(SOCKADDR_UNION *, socklen_t);
 socklen_t addr_len(const SOCKADDR_UNION *);
