@@ -41,7 +41,7 @@
 /**************************************** prototypes */
 
 NOEXPORT int name2addrlist(SOCKADDR_LIST *, char *, char *);
-NOEXPORT int hostport2addrlist(SOCKADDR_LIST *, char *, char *);
+NOEXPORT void addrlist2addr(SOCKADDR_UNION *, SOCKADDR_LIST *);
 
 #ifndef HAVE_GETADDRINFO
 
@@ -92,7 +92,8 @@ void resolver_init() {
 #if defined(USE_WIN32) && !defined(_WIN32_WCE)
     HINSTANCE handle;
 
-    handle=LoadLibrary("ws2_32.dll"); /* IPv6 in Windows XP or higher */
+        /* IPv6 in Windows XP or higher */
+    handle=LoadLibrary(TEXT("ws2_32.dll"));
     if(handle) {
         s_getaddrinfo=(GETADDRINFO)GetProcAddress(handle, "getaddrinfo");
         s_freeaddrinfo=(FREEADDRINFO)GetProcAddress(handle, "freeaddrinfo");
@@ -101,7 +102,9 @@ void resolver_init() {
             return; /* IPv6 detected -> OK */
         FreeLibrary(handle);
     }
-    handle=LoadLibrary("wship6.dll"); /* experimental IPv6 for Windows 2000 */
+
+        /* experimental IPv6 for Windows 2000 */
+    handle=LoadLibrary(TEXT("wship6.dll"));
     if(handle) {
         s_getaddrinfo=(GETADDRINFO)GetProcAddress(handle, "getaddrinfo");
         s_freeaddrinfo=(FREEADDRINFO)GetProcAddress(handle, "freeaddrinfo");
@@ -110,6 +113,8 @@ void resolver_init() {
             return; /* IPv6 detected -> OK */
         FreeLibrary(handle);
     }
+
+        /* fall back to the built-in emulation */
     s_getaddrinfo=NULL;
     s_freeaddrinfo=NULL;
     s_getnameinfo=NULL;
@@ -119,33 +124,56 @@ void resolver_init() {
 /**************************************** stunnel resolver API */
 
 int name2addr(SOCKADDR_UNION *addr, char *name, char *default_host) {
-    SOCKADDR_LIST addr_list;
+    SOCKADDR_LIST *addr_list;
     int retval;
 
-    addr_list.num=0;
-    addr_list.addr=NULL;
-    retval=name2addrlist(&addr_list, name, default_host);
-    if(retval>0)
-        memcpy(addr, &addr_list.addr[0], sizeof *addr);
-    if(addr_list.addr)
-        str_free(addr_list.addr);
+    addr_list=str_alloc(sizeof(SOCKADDR_LIST));
+    addrlist_init(addr_list);
+    retval=name2addrlist(addr_list, name, default_host);
+    if(retval)
+        addrlist2addr(addr, addr_list);
+    if(addr_list->addr)
+        str_free(addr_list->addr);
+    str_free(addr_list);
     return retval;
 }
 
-int hostport2addr(SOCKADDR_UNION *addr, char *hostname, char *portname) {
-    SOCKADDR_LIST addr_list;
+int hostport2addr(SOCKADDR_UNION *addr, char *host_name, char *port_name) {
+    SOCKADDR_LIST *addr_list;
     int retval;
 
-    addr_list.num=0;
-    addr_list.addr=NULL;
-    retval=hostport2addrlist(&addr_list, hostname, portname);
-    if(retval>0)
-        memcpy(addr, &addr_list.addr[0], sizeof *addr);
-    if(addr_list.addr)
-        str_free(addr_list.addr);
+    addr_list=str_alloc(sizeof(SOCKADDR_LIST));
+    addrlist_init(addr_list);
+    retval=hostport2addrlist(addr_list, host_name, port_name);
+    if(retval)
+        addrlist2addr(addr, addr_list);
+    if(addr_list->addr)
+        str_free(addr_list->addr);
+    str_free(addr_list);
     return retval;
 }
 
+NOEXPORT void addrlist2addr(SOCKADDR_UNION *addr, SOCKADDR_LIST *addr_list) {
+    int i;
+
+    for(i=0; i<addr_list->num; ++i) { /* find the first IPv4 address */
+        if(addr_list->addr[i].in.sin_family==AF_INET) {
+            memcpy(addr, &addr_list->addr[i], sizeof(SOCKADDR_UNION));
+            return;
+        }
+    }
+#ifdef USE_IPv6
+    for(i=0; i<addr_list->num; ++i) { /* find the first IPv6 address */
+        if(addr_list->addr[i].in.sin_family==AF_INET6) {
+            memcpy(addr, &addr_list->addr[i], sizeof(SOCKADDR_UNION));
+            return;
+        }
+    }
+#endif
+    /* copy the first address resolved (curently AF_UNIX) */
+    memcpy(addr, &addr_list->addr[0], sizeof(SOCKADDR_UNION));
+}
+
 int namelist2addrlist(SOCKADDR_LIST *addr_list, NAME_LIST *name_list, char *default_host) {
     /* recursive implementation to reverse the list */
     if(!name_list)
@@ -155,13 +183,10 @@ int namelist2addrlist(SOCKADDR_LIST *addr_list, NAME_LIST *name_list, char *defa
 }
 
 NOEXPORT int name2addrlist(SOCKADDR_LIST *addr_list, char *name, char *default_host) {
-    char *tmp, *hostname, *portname;
+    char *tmp, *host_name, *port_name;
     int retval;
 
-    addr_list->rr_val=0; /* reset round-robin counter */
-    /* allow structures created with sockaddr_dup() to modify
-     * the original rr_val rather than its local copy */
-    addr_list->rr_ptr=&addr_list->rr_val;
+    addrlist_init(addr_list);
 
     /* first check if this is a UNIX socket */
 #ifdef HAVE_STRUCT_SOCKADDR_UN
@@ -179,25 +204,25 @@ NOEXPORT int name2addrlist(SOCKADDR_LIST *addr_list, char *name, char *default_h
     }
 #endif
 
-    /* set hostname and portname */
+    /* setup host_name and port_name */
     tmp=str_dup(name);
-    portname=strrchr(tmp, ':');
-    if(portname) {
-        hostname=tmp;
-        *portname++='\0';
+    port_name=strrchr(tmp, ':');
+    if(port_name) {
+        host_name=tmp;
+        *port_name++='\0';
     } else { /* no ':' - use default host IP */
-        hostname=default_host;
-        portname=tmp;
+        host_name=default_host;
+        port_name=tmp;
     }
 
     /* fill addr_list structure */
-    retval=hostport2addrlist(addr_list, hostname, portname);
+    retval=hostport2addrlist(addr_list, host_name, port_name);
     str_free(tmp);
     return retval;
 }
 
-NOEXPORT int hostport2addrlist(SOCKADDR_LIST *addr_list,
-        char *hostname, char *portname) {
+int hostport2addrlist(SOCKADDR_LIST *addr_list,
+        char *host_name, char *port_name) {
     struct addrinfo hints, *res=NULL, *cur;
     int err, retries=0;
 
@@ -210,7 +235,7 @@ NOEXPORT int hostport2addrlist(SOCKADDR_LIST *addr_list,
     hints.ai_socktype=SOCK_STREAM;
     hints.ai_protocol=IPPROTO_TCP;
     for(;;) {
-        err=getaddrinfo(hostname, portname, &hints, &res);
+        err=getaddrinfo(host_name, port_name, &hints, &res);
         if(err && res)
             freeaddrinfo(res);
         if(err!=EAI_AGAIN || ++retries>=3)
@@ -222,11 +247,11 @@ NOEXPORT int hostport2addrlist(SOCKADDR_LIST *addr_list,
     case 0:
         break; /* success */
     case EAI_SERVICE:
-        s_log(LOG_ERR, "Unknown TCP service '%s'", portname);
+        s_log(LOG_ERR, "Unknown TCP service '%s'", port_name);
         return 0; /* error */
     default:
         s_log(LOG_ERR, "Error resolving '%s': %s",
-            hostname, s_gai_strerror(err));
+            host_name, s_gai_strerror(err));
         return 0; /* error */
     }
 
@@ -246,6 +271,17 @@ NOEXPORT int hostport2addrlist(SOCKADDR_LIST *addr_list,
     return addr_list->num; /* ok - return the number of addresses */
 }
 
+void addrlist_init(SOCKADDR_LIST *addr_list) {
+    addr_list->num=0;
+    if(addr_list->addr)
+        str_free(addr_list->addr);
+    addr_list->addr=NULL;
+    addr_list->rr_val=0; /* reset round-robin counter */
+    /* allow structures created with sockaddr_dup() to modify
+     * the original rr_val rather than its local copy */
+    addr_list->rr_ptr=&addr_list->rr_val;
+}
+
 void addrlist_dup(SOCKADDR_LIST *dst, const SOCKADDR_LIST *src) {
     memcpy(dst, src, sizeof(SOCKADDR_LIST));
     if(src->addr) {
